/*
     ps 1.0b (June 7, 1998) - Lists running processes on Windows 95/NT
     Copyright (C) 1998  Yoram Last (ylast@mindless.com)
     Based on code by Amol Deshpande (amold@microsoft.com)
     from his Win32 port of zsh (which was obtained from
     ftp://ftp.blarg.net/users/amol/zsh/).

 ------------------------------------------------------------

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

-------------------------------------------------------------

 Usage: ps [-w] [-a]

  -w     Also lists window titles of main-thread windows.

  -a     Lists all processes running on the system (by default only 
         processes to which the user running 'ps' has access privileges
         will be shown). This plays a role only on NT.

--------------------------------------------------------------*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winperf.h>
#include <tlhelp32.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#define p_malloc(s) HeapAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,(s))
#define p_realloc(s,p) HeapReAlloc(GetProcessHeap(),HEAP_ZERO_MEMORY,(s),(p))
#define p_free(p) HeapFree(GetProcessHeap(),0,(p))

#define REGKEY_PERF         "software\\microsoft\\windows nt\\currentversion\\perflib"
#define REGSUBKEY_COUNTERS  "Counters"
#define PROCESS_COUNTER     "process"
#define PROCESSID_COUNTER   "id process"

typedef struct _pslist {
	DWORD pid;
	HANDLE hwnd;
	char exename[MAX_PATH];
	char title[80];
}pslist;

typedef BOOL (WINAPI *walker)(HANDLE,LPPROCESSENTRY32);
typedef HANDLE (WINAPI *create_snapshot)(DWORD,DWORD);
static walker proc32First;
static walker proc32Next;
static create_snapshot createSnapshot;

typedef DWORD (WINAPI *qsinformation)(DWORD, DWORD *, DWORD, DWORD *);
static qsinformation NtQuerySystemInformation;

typedef DWORD (*plist_proc)(void);
DWORD Win95Lister(void);
DWORD NTLister(void);
DWORD NTLister_old_style(void);

plist_proc ProcessListFunc;
pslist *processlist;
static unsigned long numprocs, g_dowindows;

DWORD gdwPlatform;

void make_err_str(unsigned int error,char *buf,int size) {

	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,
				  NULL,
				  error, 
				  MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),
				  buf,
				  size,
				  NULL);
	return;
}

BOOL CALLBACK enum_wincb(HWND hwnd,DWORD nump) {

	unsigned int i;
	DWORD pid = 0;

	if (!GetWindowThreadProcessId(hwnd,&pid))
		return TRUE;
	
	for (i =0;i < nump;i++) {
		if (processlist[i].pid == pid){
			processlist[i].hwnd = hwnd;
			processlist[i].title[0] =0;
			GetWindowText(hwnd,processlist[i].title,sizeof(processlist[i].title));
			break;
		}
	}
	return TRUE;
}

BOOL CanAccessProcess (DWORD ProcessId) {

	HANDLE hproc;

	hproc = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ,
						FALSE, ProcessId);
	if (hproc) {
		CloseHandle(hproc);
		return TRUE;
	} else
		return FALSE;
}

DWORD Win95Lister(void) {

	HANDLE hsnap;
	PROCESSENTRY32 pe;
	unsigned long nump =0;
	unsigned int i = 1;

	hsnap = createSnapshot(TH32CS_SNAPPROCESS,0);

	if (hsnap == INVALID_HANDLE_VALUE)
		return 0;

	pe.dwSize = sizeof(PROCESSENTRY32);
	if (proc32First(hsnap,&pe) ) {
		processlist = p_malloc(100*sizeof(pslist));
		if (!processlist)
			goto done;

		while(1) {
			lstrcpy(processlist[nump].exename,pe.szExeFile);
			processlist[nump].title[0] = 0;
			processlist[nump].pid = pe.th32ProcessID;

			nump++;

			if (!proc32Next(hsnap,&pe))
				break;

			if (nump >= i*100) {
				i++;
				processlist = p_realloc(processlist,100*i*sizeof(pslist));
			}
		}
	}
done:
	CloseHandle(hsnap);

	if (g_dowindows) {
		EnumWindows(enum_wincb,nump);
	}
	return nump;
}

DWORD NTLister(void) {
	
	DWORD *sysinf_buffer;
	DWORD sbSize = 0x10000;
	unsigned long nump = 0;
	unsigned int i = 1;

	sysinf_buffer = p_malloc(sbSize);
	if (!sysinf_buffer)  
		goto done;

	while (1) {
		if (!NtQuerySystemInformation(5,sysinf_buffer,sbSize,NULL)) {
			break;
		} else {
			if (*sysinf_buffer) {
				sbSize += 0x8000;
				sysinf_buffer = p_realloc(sysinf_buffer,sbSize);
			} else {
			//	fprintf(stderr,"NtQuerySystemInformation failed\n");
			//	return 0;
				return NTLister_old_style();
			}
		}
	}

	processlist = p_malloc(100*sizeof(pslist));
	if (!processlist) {  
		goto done;}
	
	while (1) {
		processlist[nump].pid = *(sysinf_buffer+17);
		if (*(sysinf_buffer+15)) {
			sprintf(processlist[nump].exename, "%S",
						*(sysinf_buffer+15));
		} else {
			lstrcpy(processlist[nump].exename,"Idle Process");
		}
		processlist[nump].title[0] = 0;

		nump++;

		if (!*sysinf_buffer)
			break;

		sysinf_buffer += (*sysinf_buffer)/sizeof(DWORD);

		if (nump >= i*100) {
			i++;
			processlist = p_realloc(processlist,100*i*sizeof(pslist));
		}
	}

done:

	if (g_dowindows) {
		EnumWindows(enum_wincb,nump);
	}

	return nump;
}

DWORD NTLister_old_style(void) {

	DWORD                        rc;
	HKEY                         hKeyNames;
	DWORD                        dwType;
	DWORD                        dwSize;
	LPBYTE                       buf = NULL;
	char                         szSubKey[1024];
	LANGID                       lid;
	LPSTR                        p;
	LPSTR                        p2;
	PPERF_DATA_BLOCK             pPerf;
	PPERF_OBJECT_TYPE            pObj;
	PPERF_INSTANCE_DEFINITION    pInst;
	PPERF_COUNTER_BLOCK          pCounter;
	PPERF_COUNTER_DEFINITION     pCounterDef;
	DWORD                        i;
	DWORD                        dwProcessIdTitle;
	DWORD                        dwProcessIdCounter;
	char                         szProcessName[MAX_PATH];
	unsigned long                numprocesses=0;



	lid = MAKELANGID( LANG_ENGLISH, SUBLANG_NEUTRAL );
	wsprintf( szSubKey, "%s\\%03x", REGKEY_PERF, lid );
	rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE, szSubKey,0,KEY_READ,&hKeyNames);

	if (rc != ERROR_SUCCESS)
		goto exit;

	rc = RegQueryValueEx( hKeyNames, REGSUBKEY_COUNTERS, NULL, &dwType,
			NULL, &dwSize);

	if (rc != ERROR_SUCCESS) 
		goto exit;

	buf =  p_malloc( dwSize );
	if (buf == NULL) 
		goto exit;

	rc = RegQueryValueEx( hKeyNames, REGSUBKEY_COUNTERS, NULL, &dwType,
			buf, &dwSize);

	if (rc != ERROR_SUCCESS) 
		goto exit;

	p = buf;
	while (*p) {
		if (p > buf) {
			for( p2=p-2; isdigit(*p2); p2--) ;
		}
		if (lstrcmpi(p, PROCESS_COUNTER) == 0) {
			for( p2=p-2; isdigit(*p2); p2--) ;
			lstrcpy( szSubKey, p2+1 );
		}
		else if (lstrcmpi(p, PROCESSID_COUNTER) == 0) {
			for( p2=p-2; isdigit(*p2); p2--) ;
			dwProcessIdTitle = atol( p2+1 );
		}
		p += (lstrlen(p) + 1);
	}

	p_free( buf );

	dwSize = 65535;

	buf = p_malloc( dwSize );
	if (buf == NULL) 
		goto exit;

	while (1) {

		rc = RegQueryValueEx( HKEY_PERFORMANCE_DATA, szSubKey, NULL,
				&dwType, buf, &dwSize);

		pPerf = (PPERF_DATA_BLOCK) buf;

		if ((rc == ERROR_SUCCESS) &&
				(dwSize > 0) &&
				(pPerf)->Signature[0] == (WCHAR)'P' &&
				(pPerf)->Signature[1] == (WCHAR)'E' &&
				(pPerf)->Signature[2] == (WCHAR)'R' &&
				(pPerf)->Signature[3] == (WCHAR)'F' )
			break;

		if (rc == ERROR_MORE_DATA) {
			dwSize += 32767;
			buf = p_realloc( buf, dwSize );
		} else {
			goto exit;
		}
	}

	//
	// set the perf_object_type pointer
	//
	pObj = (PPERF_OBJECT_TYPE) ((DWORD)pPerf + pPerf->HeaderLength);

	//
	// loop thru the performance counter definition records looking
	// for the process id counter and then save its offset
	//
	pCounterDef = (PPERF_COUNTER_DEFINITION) ((DWORD)pObj + pObj->HeaderLength);
	for (i=0; i<(DWORD)pObj->NumCounters; i++) {
		if (pCounterDef->CounterNameTitleIndex == dwProcessIdTitle) {
			dwProcessIdCounter = pCounterDef->CounterOffset;
			break;
		}
		pCounterDef++;
	}


	pInst = (PPERF_INSTANCE_DEFINITION) ((DWORD)pObj + pObj->DefinitionLength);

	numprocesses = pObj->NumInstances;

	processlist = p_malloc(numprocesses*sizeof(pslist));

	if (!processlist){
		goto exit;
	}
	for (i=0; i<numprocesses; i++) {
		//
		// pointer to the process name
		//
		p = (LPSTR) ((DWORD)pInst + pInst->NameOffset);

		//
		// convert it to ascii
		//
		rc = WideCharToMultiByte( CP_ACP,
				0,
				(LPCWSTR)p,
				-1,
				szProcessName,
				sizeof(szProcessName),
				NULL,
				NULL);

		if (!rc) {
			lstrcpy( processlist[i].exename, "unknown" );
		}
		if (!lstrcmpi(szProcessName,"_Total")){
			numprocesses--;
			continue;
		}

		pCounter = (PPERF_COUNTER_BLOCK) ((DWORD)pInst + pInst->ByteLength);
		processlist[i].pid = *((LPDWORD) ((DWORD)pCounter + dwProcessIdCounter));

		if (lstrlen(szProcessName)+4 <= sizeof(processlist[i].exename)) {
			if (processlist[i].pid == 0 && !lstrcmpi(szProcessName,"Idle")) {
				lstrcpy( processlist[i].exename, "Idle Process" );
			} else {
				lstrcpy( processlist[i].exename, szProcessName );
				if (!(processlist[i].pid == 2) || lstrcmpi(szProcessName,"System"))
					lstrcat( processlist[i].exename, ".exe" );
			}
			
		}

		pInst = (PPERF_INSTANCE_DEFINITION) 
			((DWORD)pCounter + pCounter->ByteLength);
	}

	if (g_dowindows) {
		EnumWindows(enum_wincb,numprocesses);
	}
exit:
	if (buf) {
		p_free( buf );
	}

	RegCloseKey( hKeyNames );
	RegCloseKey( HKEY_PERFORMANCE_DATA );

	return numprocesses;
}

void init_plister(void) {

	HANDLE hlib;
	char errbuf[128];

	if (gdwPlatform == VER_PLATFORM_WIN32_NT) {

		hlib = LoadLibrary("ntdll.dll");

		if (!hlib){
			ProcessListFunc = NTLister_old_style;
			return;
		}

		NtQuerySystemInformation = (qsinformation)
						GetProcAddress(hlib,"NtQuerySystemInformation");

		if (!NtQuerySystemInformation) {
			FreeLibrary(hlib);
			//fprintf(stderr,"Can't find NtQuerySystemInformation in ntdll.dll\n");
			ProcessListFunc = NTLister_old_style;
			return;
		}
		ProcessListFunc = NTLister;

	} else {
		hlib = LoadLibrary("kernel32.dll");
		if (!hlib){
			make_err_str(GetLastError(),errbuf,128);
			fprintf(stderr,"Can't load kernel32.dll:%s)\n",errbuf);
			return;
		}
		proc32First = (walker)GetProcAddress(hlib,"Process32First");
		proc32Next = (walker)GetProcAddress(hlib,"Process32Next");
		createSnapshot= (create_snapshot)GetProcAddress(hlib,
								"CreateToolhelp32Snapshot");
		if (!proc32First || !proc32Next || !createSnapshot) {
			FreeLibrary(hlib);
			fprintf(stderr,"Can't find exports in kernel32.dll\n");
			return;
		}
		ProcessListFunc = Win95Lister;
	}

}

int Usage() {

	fprintf( stderr, "ps 1.0b (June 7, 1998) - Lists running processes on Windows 95/NT\n");
	fprintf( stderr, "Copyright (C) 1998 Yoram Last (ylast@mindless.com)\n");
	fprintf( stderr, "This program is free software distributed under the terms of the GNU\n");
	fprintf( stderr, "General Public Licence. This program comes with ABSOLUTELY NO WARRANTY.\n");
	fprintf( stderr, "See the accompanying documentation for more information.\n\n" );
	fprintf( stderr, "Usage: ps [-w] [-a]\n\n" );
	fprintf( stderr, "  -w        Also lists window titles of main-thread windows.\n\n");
	fprintf( stderr, "  -a        Lists all processes running on the system (by default only\n");
	fprintf( stderr, "            processes to which the user running 'ps' has access privileges\n");
	fprintf( stderr, "            will be shown). This plays a role only on NT.\n");
	return 0;
}

int main (unsigned int argc, char **argv) {

	OSVERSIONINFO osvi;
	DWORD nump;
	BOOL showall = FALSE;
	unsigned int i,k;

	memset(&osvi,0,sizeof(OSVERSIONINFO));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osvi);
	gdwPlatform = osvi.dwPlatformId;

	g_dowindows = 0;

	if (argc > 1) {
		k = 1;
		while (k < argc && (!showall || !g_dowindows)) {
			if (argv[k][0] == '-' || argv[k][0] == '/') {
				i = 1;
				while (argv[k][i] && (!showall || !g_dowindows)) {
					if (argv[k][i] == 'w' || argv[k][i] == 'W')
						g_dowindows = 1;
					if (argv[k][i] == 'a' || argv[k][i] == 'A')
						showall = TRUE;
					i++;
				}
			}
			k++;
		}
		if (!showall && !g_dowindows)
				return Usage();
	}

	init_plister();
	if (!ProcessListFunc)
		return 1;
	nump = ProcessListFunc();

	for (i=0; i< nump; i++) {
		if (g_dowindows) {
			if (gdwPlatform == VER_PLATFORM_WIN32_NT) {
				if (showall || CanAccessProcess(processlist[i].pid))
					printf("%-5u  %-20s    %s\n",processlist[i].pid,
									processlist[i].exename,
									processlist[i].title);
			} else {
				printf("%-10d  %-40s    %s\n",processlist[i].pid,
								processlist[i].exename,
								processlist[i].title);
			}
		} else {
			if (gdwPlatform == VER_PLATFORM_WIN32_NT) {
				if (showall || CanAccessProcess(processlist[i].pid))
					printf("%-5u  %s\n",processlist[i].pid,
									processlist[i].exename);
			} else {
				printf("%-10d  %s\n",processlist[i].pid,
								processlist[i].exename);
			}
		}
	}
	return 0;
}
