# fslib-win32.pl - File manipulation subroutines
# This file is part of WebRSH 1.1b (beta)
# Copyright (C) 1997,1998  Yoram Last (ylast@mindless.com)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

use Win32::File;

sub MakeDir { # Creates a new directory.

  local($dirname) = @_;
  mkdir $dirname, 0777;
}

sub DelTree { # Deletes a file or a directory tree.

  local($file) = @_;
  if (-d $file) {
    &RmDirTree($file) || &WebrshDie("Can\'t remove $file \: $!");
  } else {
    Win32::File::SetAttributes($file, 128);
    unlink($file) || &WebrshDie("Can\'t delete $file \: $!");
  };
  return 1;
}

sub RmDirTree { # Deletes a directory tree.

  local($dir) = @_;
  opendir(DIR,$dir) || &WebrshDie("Can\'t open $dir \: $!");
  local(@items) = grep !/^\.\.?$/, readdir DIR;
  close DIR;
  local($i);
  foreach $i (0..$#items) {
    if (-d "$dir/$items[$i]" && $items[$i] ne '..') {
      &RmDirTree("$dir/$items[$i]") || &WebrshDie("Can\'t remove $dir/$items[$i] \: $!");
    } else {
      Win32::File::SetAttributes("$dir/$items[$i]", 128);
      unlink("$dir/$items[$i]") || &WebrshDie("Can\'t remove $dir/$items[$i] \: $!");
    };
  };
  Win32::File::SetAttributes($dir, 16);
  rmdir "$dir";
}

sub FCopy { # Copies a file.

  local($file,$dest) = @_;
  local($buf) = '';
  $dest =~ s/[\\\/]$//;
  if (-d $dest) {
    $dest = join('/',$dest,&FileName($file));
  };
  open (FILE,"<$file") || &WebrshDie("Can\'t open $file: $!");
  binmode FILE;
  open (DEST,">$dest") || &WebrshDie("Can\'t write $dest: $!");
  binmode DEST;
  while (read(FILE,$buf,$BufSize)) {print DEST $buf;};
  close FILE; close DEST;
  chmod int((stat $file)[2]), $dest;
  local($Attributes);
  Win32::File::GetAttributes($file, $Attributes);
  Win32::File::SetAttributes($dest, $Attributes);
  return -e $dest;
}

sub DCopy { # Copies a directory tree.

  local($dir,$dest) = @_;
  $dest =~ s/[\\\/]$//;
  $dir =~ s/[\\\/]$//;
  -e $dir || &WebrshDie("Can\'t find $dir \: $!");
  -d $dir || &WebrshDie("$dir is not a directory");
  if (-e $dest) {
    unless (-d $dest) {
      &WebrshDie("Can\'t copy $dir to $dest \: Destination exists and is not a directory");
    };
  } else {
    mkdir $dest, 0600;
    chmod 0600, $dest;
  };
  opendir(DIR,$dir) || &WebrshDie("Can\'t open $dir \: $!");
  local(@items) = grep !/^\.\.?$/, readdir DIR;
  close DIR;
  local($i);
  foreach $i (0..$#items) {
    if (-d "$dir/$items[$i]" && $items[$i] ne '..') {
      &DCopy("$dir/$items[$i]","$dest/$items[$i]") || 
        &WebrshDie("Failed to copy $dir/$items[$i] to $dest/$items[$i] \: $!");
    } else {
      &FCopy("$dir/$items[$i]","$dest/$items[$i]") ||
        &WebrshDie("Failed to copy $dir/$items[$i] to $dest/$items[$i] \: $!");
    };
  };
  chmod int((stat $dir)[2]), $dest;
  local($Attributes);
  Win32::File::GetAttributes($file, $Attributes);
  Win32::File::SetAttributes($dest, $Attributes);
  return -e $dest;
}

sub GetAttributes {

  local($file) = @_;
  local($atA,$atS,$atH,$atR);
  local($Att);
  if (!$file || $file eq '.') {
    $file = $pwd;
  } elsif ($file eq '..') {
    $file = $pwd;
    $file =~ s/[^\\^\/]+$//;
    chop($file);
  };
  Win32::File::GetAttributes($file, $Att);
  $Att = sprintf "%o", int($Att);
  if ($Att < 8) {
    ($atR,$atH,$atS) = split(//,unpack "b3", $Att);
    $atA = 0;
  } else {
    local(@Attd) = split(//,substr("$Att",-2,2));
    ($atR,$atH,$atS) = split(//,unpack "b3", @Attd[1]);
    $atA = (split(//,unpack "b3", @Attd[0]))[2];
  };
  return ($atA,$atS,$atH,$atR);
}

sub SetAttributes { # Sets the attributes of a file or directory.

  local($file,$atA,$atS,$atH,$atR) = @_;
  local($Att,$OAtt,$OatA,$OatS,$OatH,$OatR);
  if (!$file || $file eq '.') {
    $file = $pwd;
  } elsif ($file eq '..') {
    $file = $pwd;
    $file =~ s/[^\\^\/]+$//;
    chop($file);
  };
  Win32::File::GetAttributes($file, $Att);
  $OAtt = sprintf "%o", int($Att);
  if ($OAtt < 8) {
    ($OatR,$OatH,$OatS) = split(//,unpack "b3", $OAtt);
    $OatA = 0;
  } else {
    local(@OAttd) = split(//,substr("$OAtt",-2,2));
    ($OatR,$OatH,$OatS) = split(//,unpack "b3", @OAttd[1]);
    $OatA = (split(//,unpack "b3", @OAttd[0]))[2];
  };
  $Att = $Att + 32*($atA - $OatA) + 4*($atS - $OatS) + 2*($atH - $OatH) + $atR - $OatR;
  if ($Att > 128) {$Att = $Att - 128;};
  Win32::File::SetAttributes($file, $Att);
  return 1;
}

sub GetInfo { # Gets "dir" information for a file or a directory.

  local($file) = @_;
  $file =~ s/\%/\%\%/g if ($OStype eq 'w95');
  $file =~ s/\//\\/g;
  $_ = &ExWin32Com("dir /A $file");
  return $_;
}

sub PrintAttForm { # Prints a form for viewing and setting Win32 file attributes.

  local($file) = @_;
  local($atA,$atS,$atH,$atR) = &GetAttributes($file);
  print '<FORM METHOD="',"$ReqMeth",'" ACTION="',"$ScriptUrl",'">',"\n",
  '<INPUT TYPE=hidden name="id" value="',"$id",'">',"\n",
  '<INPUT TYPE=hidden name="pwd" value="',"$pwd",'">',"\n",
  '<INPUT TYPE=hidden name="com" value="rshfm">',"\n",
  '<INPUT TYPE=hidden name="file" value="',"$file",'">',"\n",
  '<INPUT TYPE=hidden name="props" value="1">',"\n",
  '<TABLE><TR><TD COLSPAN=32><B> Attributes: </B></TD><TD COLSPAN=24>',"\n",
  '<INPUT TYPE=checkbox name="atA" value="1"'; if ($atA) {print ' CHECKED';}; 
  print '>Archive </TD><TD COLSPAN=28>',"\n",
  '<INPUT TYPE=checkbox name="atS" value="1"'; if ($atS) {print ' CHECKED';}; 
  print '>System </TD><TD COLSPAN=28>',"\n",
  '<INPUT TYPE=checkbox name="atH" value="1"'; if ($atH) {print ' CHECKED';}; 
  print '>Hidden </TD><TD COLSPAN=28>',"\n",
  '<INPUT TYPE=checkbox name="atR" value="1"'; if ($atR) {print ' CHECKED';}; 
  print '>Read-only </TD></TR></TABLE><BR>',"\n",
  '<INPUT TYPE=submit name="fmcom" VALUE="     OK      ">',"\n",
  '<INPUT TYPE=submit VALUE="    Cancel   ">',"\n",
  '<INPUT TYPE=submit name="fmcom" VALUE="    Apply    "></FORM>',"\n";
}

1; # return true
